/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2019, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2019-03-27
 */

#include "T2TSearch.h"
#include "T2TUtility.h"
#include "../../tensor/core/CHeader.h"

using namespace nts;

namespace transformer
{

/*
initialize the model
>> argc - number of arguments
>> argv - list of pointers to the arguments
*/
void T2TSearch::InitModel(int argc, char ** argv)
{
    LoadParamInt(argc, argv, "beamsize", &beamSize, 1);
}

/* 
search for the most promising states 
>> model - the transformer model
>> input - input of the model
>> padding - padding of the input
>> output - output that represents the sequences as rows
*/
void T2TSearch::Search(T2TModel * model, XTensor * input, XTensor * padding, XTensor * output)
{
    XTensor maskEnc;
    XTensor encoding;
    T2TPredictor predictor;

    /* encoder mask */
    model->MakeMTMaskEnc(*input, *padding, maskEnc);
    
    /* make the encoding network */
    encoding = model->MakeEncoder(*input, maskEnc, false);
    encoding.SetName(ENCODING_NAME);
    
    T2TStateBundle * states = new T2TStateBundle[maxLength];
    T2TStateBundle * first = states;
    
    /* create the first state */
    predictor.Create(model, &encoding, first);

    /* generate the sequence from left to right */
    for(int i = 0 ; i < maxLength; i++){
        T2TStateBundle * cur = states + i;
        T2TStateBundle * next = states + i + 1;

        /* read the current state */
        predictor.Read(model, cur);

        /* predict the next state */
        predictor.Predict(next, &encoding, input, padding);

        /* beam pruning */
        Generate(next);
    }
    
    delete[] states;
}

/* 
generate tokens for the next state via beam pruning
>> beam - the beam that keeps a number of states
*/
void T2TSearch::Generate(T2TStateBundle * beam)
{
    int dims[MAX_TENSOR_DIM_NUM];
    int dimsBeam[MAX_TENSOR_DIM_NUM];
    int dimsTopK[MAX_TENSOR_DIM_NUM];
    
    XTensor scoreTopK;
    XTensor &score = beam->score;
    XTensor &index = beam->prediction;
    XTensor &preID = beam->preID;
    int order = score.order;

    CheckNTErrors(order >= 2, "The tensor must be of order 2 or larger.");
    CheckNTErrors(dimsBeam[order - 2] % beamSize == 0, "Wrong dimension size!");
    
    for (int i = 0; i < order; i++) {
        dims[i] = score.GetDim(i);
        dimsBeam[i] = score.GetDim(i);
        dimsTopK[i] = score.GetDim(i);
    }

    dimsBeam[order - 2] /= beamSize;
    dimsBeam[order - 1] *= beamSize;
    dimsTopK[order - 2] =  dimsBeam[order - 2];
    dimsTopK[order - 1] =  beamSize;
    
    InitTensor(&scoreTopK, order, dimsTopK, score.dataType,
               1.0F, score.devID, score.mem);
    InitTensor(&index, order, dimsTopK, X_INT,
               1.0F, score.devID, score.mem);
    InitTensor(&preID, order, dimsTopK, X_INT,
               1.0F, -1);
    
    score.Reshape(order, dimsBeam);
    
    /* keep the most promissing candidates in the beam */
    TopK(score, scoreTopK, index, 0, beamSize);
    
    CopyValues(scoreTopK, preID);
    
    int sizePredict = score.GetDim(-1);
    
    /* pre id !!! */
    Descale(preID, sizePredict);
    
    /* mod !!! */

    score.Reshape(order, dims);
}

/* 
expand the search graph 
>> beam - the beam that keeps a number of states
*/
void T2TSearch::Expand(T2TStateBundle * prev, T2TStateBundle * beam)
{
    beam->MakeStates(beam->prediction.unitNum);

    T2TState * states = beam->states;
    XTensor &predict = beam->prediction;
    
    XTensor index = *NewTensorBuf(predict.order - 1, predict.dimSize, X_FLOAT, 1.0F, 
                                  predict.devID, predict.mem);

    index.SetAscendingOrder(-1);

    DelTensorBuf(&index);
}

/* 
save the output sequences in a tensor 
>> beam - the beam that keeps a number of states
*/
void T2TSearch::DumpOutput(T2TStateBundle * beam, XTensor * output)
{
}

}
